"use strict";
var Promise = require('bluebird');
var intersection = require('lodash/intersection');
var assign = require('lodash/assign');
function buildSynchronizer(spec, options) {
    var pool = null;
    function synchronizer(transaction, item, errors) {
        if (item.action === 'INSERT') {
            return insert(pool, spec.name, options && options.insertErrorMessageBuilder, options && options.insertInterceptor, transaction, item, errors);
        }
        if (item.action === 'UPDATE') {
            return update(pool, spec.name, options && options.updateErrorMessageBuilder, options && options.updateInterceptor, transaction, item, errors);
        }
        if (item.action === 'REMOVE') {
            return remove(pool, spec.name, options && options.removeErrorMessageBuilder, options && options.removeInterceptor, transaction, item, errors);
        }
        return Promise.reject(new Error('provided item is invalid'));
    }
    function initialize(p) {
        pool = p;
    }
    return { synchronizer: synchronizer, initialize: initialize };
}
exports.buildSynchronizer = buildSynchronizer;
var INJECTED_MASTER_RESOLUTION_KEY = '__INJECTED_MASTER_RESOLUTION_KEY';
function insert(pool, model, messageBuilder, interceptor, transaction, item, errors) {
    function defaultInsertBehavior() {
        var data = assign({}, item.data);
        var masters = [];
        pool.models[model].spec.relations.master.forEach(function (master) {
            var masterData = data[master.alias];
            if (masterData) {
                var masterModel = pool.models[master.model];
                masters.push(masterModel.synchronizer(transaction, (_a = {
                        action: 'INSERT',
                        itemId: item.itemId,
                        model: masterModel.name,
                        data: masterData
                    },
                    _a[INJECTED_MASTER_RESOLUTION_KEY] = function (id) { return data[master.key] = id; },
                    _a
                ), errors));
                data[master.alias] = null;
            }
            var _a;
        });
        var modelCreationPromise = Promise.all(masters).then(function () {
            return pool.models[model].sequelizeModel.create(data, {
                transaction: transaction,
                fields: pool.models[model].allowedInsertAttributes
            });
        });
        var promise = modelCreationPromise.then(function (instance) {
            if (item[INJECTED_MASTER_RESOLUTION_KEY]) {
                item[INJECTED_MASTER_RESOLUTION_KEY](instance[pool.models[model].primaryKey]);
            }
            var details = [];
            var aliasModelMap = pool.models[model].aliasModelMap;
            var aliasKeyMap = pool.models[model].aliasKeyMap;
            Object.keys(aliasModelMap).forEach(function (alias) {
                var nested = data[alias];
                if (nested) {
                    if (Array.isArray(nested)) {
                        nested.forEach(function (nestedItem) {
                            details.push(pool.models[aliasModelMap[alias]].synchronizer(transaction, {
                                action: 'INSERT',
                                itemId: item.itemId,
                                model: aliasModelMap[alias],
                                data: assign({}, nestedItem, (_a = {}, _a[aliasKeyMap[alias]] = instance[pool.models[model].primaryKey], _a))
                            }, errors));
                            var _a;
                        });
                    }
                    else {
                        details.push(pool.models[aliasModelMap[alias]].synchronizer(transaction, {
                            action: 'INSERT',
                            itemId: item.itemId,
                            model: aliasModelMap[alias],
                            data: assign({}, nested, (_a = {}, _a[aliasKeyMap[alias]] = instance[pool.models[model].primaryKey], _a))
                        }, errors));
                    }
                }
                var _a;
            });
            return details.length > 0 ? Promise.all(details) : undefined;
        });
        if (typeof messageBuilder === 'function') {
            return promise.catch(function (error) { return errors.push({ itemId: item.itemId, message: messageBuilder(item, error) }); });
        }
        return promise;
    }
    if (typeof interceptor === 'function') {
        return interceptor(transaction, item, errors, pool).then(function (resume) {
            if (resume) {
                return defaultInsertBehavior();
            }
        });
    }
    return defaultInsertBehavior();
}
function update(pool, model, messageBuilder, interceptor, transaction, item, errors) {
    function defaultUpdateBehavior() {
        var primaryKey = pool.models[model].primaryKey;
        var promise = pool.models[model].sequelizeModel.update(item.data, {
            transaction: transaction,
            validate: true,
            fields: buildUpdateAttributes(pool.models[model].allowedUpdateAttributes, item.attributes),
            where: (_a = {}, _a[primaryKey] = item.data[primaryKey], _a)
        });
        if (typeof messageBuilder === 'function') {
            return promise.catch(function (error) { return errors.push({ itemId: item.itemId, message: messageBuilder(item, error) }); });
        }
        return promise;
        var _a;
    }
    if (typeof interceptor === 'function') {
        return interceptor(transaction, item, errors, pool).then(function (resume) {
            if (resume) {
                return defaultUpdateBehavior();
            }
        });
    }
    return defaultUpdateBehavior();
}
function remove(pool, model, messageBuilder, interceptor, transaction, item, errors) {
    function defaultRemoveBehavior() {
        var promise = pool.models[model].sequelizeModel.destroy({
            transaction: transaction,
            where: { id: item.targetId }
        });
        if (typeof messageBuilder === 'function') {
            return promise.catch(function (error) { return errors.push({ itemId: item.itemId, message: messageBuilder(item, error) }); });
        }
        return promise;
    }
    if (typeof interceptor === 'function') {
        return interceptor(transaction, item, errors, pool).then(function (resume) {
            if (resume) {
                return defaultRemoveBehavior();
            }
        });
    }
    return defaultRemoveBehavior();
}
function buildUpdateAttributes(allowedAttributes, updateAttributes) {
    if (Array.isArray(allowedAttributes)) {
        return intersection(allowedAttributes, updateAttributes);
    }
    return updateAttributes;
}
